import numpy as np

from typing import List, Dict, Tuple, Optional

from dover_lap.libs.utils import groupby
from dover_lap.libs.turn import Turn, merge_turns
from spyder import DER

from .map_utils import *


class HungarianMap:
    def __init__(self) -> None:
        pass

    def compute_mapping(
        self,
        turns_list: List[List[Turn]],
    ) -> Tuple[Dict[Tuple[int, str], int], np.ndarray]:
        """
        Use Hungarian algorithm for label mapping for 2 system special case.
        """
        self.turns_list = turns_list

        weights = self._compute_weights()
        self.sorted_turns_list = self.turns_list
        # Sort the hypotheses by their weights
        sorted_idx = weights.argsort().tolist()
        self.sorted_turns_list = [self.turns_list[i] for i in sorted_idx]

        cur_turns = self.sorted_turns_list[0]
        self.global_mapping = dict()

        for i in range(1, len(self.sorted_turns_list)):
            next_turns = self.sorted_turns_list[i]
            local_mapping = self._map_pair(cur_turns, next_turns)
            cur_turns = self._merge_pair(cur_turns, next_turns, local_mapping)
            self._update_global_map(local_mapping)

        if not self._validate_global_mapping():
            raise Exception("Some speakers have not been mapped")
        return self.global_mapping, weights

    def _validate_global_mapping(self) -> bool:
        for i, turns in enumerate(self.sorted_turns_list):
            groups = {
                key: list(group)
                for key, group in groupby(turns, lambda x: x.speaker_id)
            }
            for spk in groups:
                if (i, spk) not in self.global_mapping:
                    return False
        return True

    def _compute_weights(self) -> np.ndarray:
        N = len(self.turns_list)
        DERs = np.zeros((N, N))
        for i in range(N):
            ref = [
                (turn.speaker_id, turn.onset, turn.offset)
                for turn in self.turns_list[i]
            ]
            for j in range(N):
                if i == j:
                    continue
                hyp = [
                    (turn.speaker_id, turn.onset, turn.offset)
                    for turn in self.turns_list[j]
                ]
                DERs[i, j] = DER(ref, hyp).der
        mean_ders = DERs.mean(axis=1)
        return -1 * mean_ders

    def _map_pair(
        self, ref_turns: List[Turn], sys_turns: List[Turn]
    ) -> Dict[Tuple[int, str], int]:
        ref = [(turn.speaker_id, turn.onset, turn.offset) for turn in ref_turns]
        sys = [(turn.speaker_id, turn.onset, turn.offset) for turn in sys_turns]
        metrics = DER(ref, sys)
        ref_map = metrics.ref_map
        sys_map = metrics.hyp_map

        label_mapping = {}
        for k, v in ref_map.items():
            label_mapping[(0, k)] = v
        for k, v in sys_map.items():
            label_mapping[(1, k)] = v

        return label_mapping

    def _merge_pair(
        self,
        ref_turns: List[Turn],
        sys_turns: List[Turn],
        label_map: Dict[Tuple[int, int], int],
    ) -> List[Turn]:
        ref_turns_mapped = [
            Turn(
                onset=turn.onset,
                offset=turn.offset,
                speaker_id=label_map[(0, turn.speaker_id)],
            )
            for turn in ref_turns
        ]
        sys_turns_mapped = [
            Turn(
                onset=turn.onset,
                offset=turn.offset,
                speaker_id=label_map[(1, turn.speaker_id)],
            )
            for turn in sys_turns
        ]
        all_turns = merge_turns(ref_turns_mapped + sys_turns_mapped)
        return all_turns

    def _update_global_map(self, local_map: Dict[Tuple[int, str], int]) -> None:
        if not self.global_mapping:
            self.global_mapping = local_map.copy()
            return
        new_global_map = {}
        max_file_id = 0
        for key, old_id in self.global_mapping.items():
            file_id, spk_id = key
            max_file_id = max(max_file_id, file_id)
            new_global_map[key] = local_map[(0, old_id)]
        for key, val in local_map.items():
            file_id, spk_id = key
            if file_id == 1:
                new_global_map[(max_file_id + 1, spk_id)] = val
        self.global_mapping = new_global_map.copy()



"""Classes for representing speaker turns and interacting with RTTM files.

Taken from https://github.com/nryant/dscore
"""
from collections import namedtuple
from intervaltree import Interval, IntervalTree

from .uem import UEM
from .utils import groupby, warn, xor


Token = namedtuple("Token", ["type", "timestamp", "speaker", "hyp_index", "weight"])


class Turn(object):
    """Speaker turn class.

    A turn represents a segment of audio attributed to a single speaker.

    Parameters
    ----------
    onset : float
        Onset of turn in seconds from beginning of recording.

    offset : float, optional
        Offset of turn in seconds from beginning of recording. If None, then
        computed from ``onset`` and ``dur``.
        (Default: None)

    dur : float, optional
        Duration of turn in seconds. If None, then computed from ``onset`` and
        ``offset``.
        (Default: None)

    speaker_id : str, optional
        Speaker id.
        (Default: None)

    file_id : str, optional
        File id.
        (Default: none)
    """

    def __init__(
        self, onset, offset=None, dur=None, speaker_id=None, file_id=None, weight=None
    ):
        if not xor(offset is None, dur is None):
            raise ValueError("Exactly one of offset or dur must be given")
        if onset < 0:
            raise ValueError("Turn onset must be >= 0 seconds")
        if offset:
            dur = offset - onset
        if dur <= 0:
            raise ValueError("Turn duration must be > 0 seconds")
        if not offset:
            offset = onset + dur
        self.onset = onset
        self.offset = offset
        self.dur = dur
        self.speaker_id = speaker_id
        self.file_id = file_id
        self.weight = weight

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.onset, self.offset, self.dur, self.file_id, self.speaker_id))

    def __str__(self):
        return "FILE: %s, SPEAKER: %s, ONSET: %f, OFFSET: %f, DUR: %f" % (
            self.file_id,
            self.speaker_id,
            self.onset,
            self.offset,
            self.dur,
        )

    def __repr__(self):
        speaker_id = "'%s'" % self.speaker_id if self.speaker_id is not None else None
        file_id = "'%s'" % self.file_id if self.file_id is not None else None
        return "Turn(%f, %f, None, %s, %s)" % (
            self.onset,
            self.offset,
            speaker_id,
            file_id,
        )


def merge_turns(turns):
    """Merge overlapping turns by same speaker within each file."""
    # Merge separately within each file and for each speaker.
    new_turns = []
    for (file_id, speaker_id), speaker_turns in groupby(
        turns, lambda x: (x.file_id, x.speaker_id)
    ):
        speaker_turns = list(speaker_turns)
        speaker_it = IntervalTree.from_tuples(
            [(turn.onset, turn.offset) for turn in speaker_turns]
        )
        n_turns_pre = len(speaker_it)
        speaker_it.merge_overlaps(strict=False)
        n_turns_post = len(speaker_it)
        if n_turns_post < n_turns_pre:
            speaker_turns = []
            for intrvl in speaker_it:
                speaker_turns.append(
                    Turn(
                        intrvl.begin, intrvl.end, speaker_id=speaker_id, file_id=file_id
                    )
                )
            speaker_turns = sorted(speaker_turns, key=lambda x: (x.onset, x.offset))
            # warn('Merging overlapping speaker turns. '
            #      'FILE: %s, SPEAKER: %s' % (file_id, speaker_id))
        new_turns.extend(speaker_turns)
    return new_turns


def chop_tree(tree, onset, offset):
    """Trim Intervals so that none overlap [``onset``, ``offset``].

    Intervals contained entirely within the chopped region are removed and
    those overlapping, but not contained are trimmed back. Differs from
    ``IntervalTree.chop`` in that it keeps track of which intervals in the
    tree were affected.

    This is an inplace operation.

    Parameters
    ----------
    tree : IntervalTree
        Interval tree.

    onset : float
        Onset of chopped region.

    offset : float
        Offset of chopped region.

    Returns
    -------
    affected_intervals : set of Interval
        Intervals from ``tree`` that overlap chopped region.
    """
    overlapped_intervals = set()  # Intervals overlapping chopped region.
    insertions = set()  # Intervals to add.

    # Identify intervals contained entirely within [onset, offset].
    overlapped_intervals.update(tree.envelop(onset, offset))

    # Identify all other intervals overlapping [onset, offset]. These belong
    # to two classes:
    # - right overlap  --  interval.begin < onset
    # - left overlap  --  offset < interval.end
    for intrvl in tree.at(onset):
        if intrvl.begin >= onset:
            continue
        overlap_dur = intrvl.end - onset
        if not overlap_dur:
            continue
        overlapped_intervals.add(intrvl)
        insertions.add(Interval(intrvl.begin, onset, intrvl.data))
    for intrvl in tree.at(offset):
        if intrvl.end <= offset:
            continue
        overlap_dur = offset - intrvl.begin
        if not overlap_dur:
            continue
        overlapped_intervals.add(intrvl)
        insertions.add(Interval(offset, intrvl.end, intrvl.data))

    # Update tree.
    for intrvl in overlapped_intervals:
        tree.discard(intrvl)
    tree.update(insertions)

    return overlapped_intervals


MAX_SESSION_DUR = 1e6  # Maximum duration (seconds) of session. Any outlandishly
# high number will do.


def trim_turns(turns, uem=None, score_onset=None, score_offset=None):
    """Trim turns to scoring regions defined in UEM.

    Parameters
    ----------
    turns : list of Turn
        Speaker turns.

    uem : UEM, optional
        Un-partitioned evaluation map.
        (Default: None)

    score_onset : float, optional
        Onset of scoring region in seconds from beginning of file. Only valid
        if ``uem=None``.
        (Default: None)

    score_offset : float, optional
        Offset of scoring region in seconds from beginning of file. Only
        valid if ``uem=None``.
        (Default: None)

    Returns
    -------
    trimmed_turns : list of Turn
        Trimmed turns.
    """
    # Validate arguments.
    if uem is not None:
        if not (score_onset is None and score_offset is None):
            raise ValueError(
                "Either uem or score_onset and score_offset must " "be specified."
            )
    else:
        if score_onset is None or score_offset is None:
            raise ValueError(
                "Either uem or score_onset and score_offset must " "be specified."
            )
        if score_onset < 0:
            raise ValueError("Scoring region onset must be >= 0 seconds")
        if score_offset <= score_onset:
            raise ValueError("Scoring region duration must be > 0 seconds")

    # If no UEM provided, set each file to have same scoring region:
    # (score_onset, score_offset).
    if uem is None:
        file_ids = set([turn.file_id for turn in turns])
        uem = UEM({fid: [(score_onset, score_offset)] for fid in file_ids})

    # Trim turns to scoring regions.
    new_turns = []
    for file_id, file_turns in groupby(turns, lambda x: x.file_id):
        if file_id not in uem:
            for turn in file_turns:
                warn("Skipping turn from file not in UEM. TURN: %s" % turn)
            continue

        # Remove overlaps with no score regions.
        noscore_tree = IntervalTree.from_tuples([(0.0, MAX_SESSION_DUR)])
        for score_onset, score_offset in uem[file_id]:
            noscore_tree.chop(score_onset, score_offset)
        turns_tree = IntervalTree.from_tuples(
            (turn.onset, turn.offset, turn) for turn in file_turns
        )
        overlapped_turns = set()  # Turns found to overlap a no score region.
        for noscore_intrvl in noscore_tree:
            overlapped_intrvls = chop_tree(
                turns_tree, noscore_intrvl.begin, noscore_intrvl.end
            )
            overlapped_turns.update([intrvl.data for intrvl in overlapped_intrvls])

        # Convert interval tree to turns.
        for intrvl in turns_tree:
            orig_turn = intrvl.data
            new_turns.append(
                Turn(
                    intrvl.begin,
                    intrvl.end,
                    speaker_id=orig_turn.speaker_id,
                    file_id=orig_turn.file_id,
                )
            )

        # Report any overlapping turns to STDERR.
        for turn in sorted(overlapped_turns, key=lambda x: (x.onset, x.offset)):
            warn("Truncating turn overlapping non-scoring region. TURN: %s" % turn)

    return new_turns



"""Functions for reading/writing and manipulating NIST un-partitioned
evaluation maps.

An un-partitioned evaluation map (UEM) specifies the time regions within each
file that will be scored.

Taken from https://github.com/nryant/dscore
"""
from collections import defaultdict
from collections.abc import MutableMapping

import os

from intervaltree import IntervalTree


class UEM(MutableMapping):
    """Un-partitioned evaluation map (UEM).

    A UEM defines a mapping from file ids to scoring regions.
    """

    def __init__(self, *args, **kwargs):
        super(UEM, self).__init__()
        self.update(*args, **kwargs)

    def __setitem__(self, fid, score_regions):
        # Validate types. Expects sequence of (onset, offset) pairs.
        invalid_type_msg = "Expected sequence of pairs. Received: %r (%s)." % (
            score_regions,
            type(score_regions),
        )
        try:
            score_regions = [tuple(region) for region in score_regions]
        except TypeError:
            raise TypeError(invalid_type_msg)
        for score_region in score_regions:
            if len(score_region) != 2:
                raise TypeError(invalid_type_msg)

        # Validate that the (onset, offset) pairs are valid: no negative
        # timestamps or negative durations.
        def _convert_to_float(score_region):
            onset, offset = score_region
            try:
                onset = float(onset)
                offset = float(offset)
            except ValueError:
                raise ValueError(
                    "Could not convert interval onset/offset to float: %s"
                    % repr(score_region)
                )
            if onset >= offset or onset < 0:
                raise ValueError(
                    'Invalid interval (%.3f, %.3f) for file "%s".'
                    % (onset, offset, fid)
                )
            return onset, offset

        score_regions = [_convert_to_float(region) for region in score_regions]

        # Merge overlaps. Use of intervaltree Incurs some additional overhead,
        # but pretty small compared to actual scoring.
        tree = IntervalTree.from_tuples(score_regions)
        tree.merge_overlaps()
        score_regions = [(intrvl.begin, intrvl.end) for intrvl in tree]

        self.__dict__[fid] = score_regions

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return "{}, UEM({})".format(super(UEM, self).__repr__(), self.__dict__)


def load_uem(uemf):
    """Load un-partitioned evaluation map from file in NIST format.

    The un-partitioned evaluation map (UEM) file format contains
    one record per line, each line consisting of NN space-delimited
    fields:

    - file id  --  file id
    - channel  --  channel (1-indexed)
    - onset  --  onset of evaluation region in seconds from beginning of file
    - offset  --  offset of evaluation region in seconds from beginning of
      file

    Lines beginning with semicolons are regarded as comments and ignored.

    Parameters
    ----------
    uemf : str
        Path to UEM file.

    Returns
    -------
    uem : UEM
        Evaluation map.
    """
    with open(uemf, "rb") as f:
        fid_to_score_regions = defaultdict(list)
        for line in f:
            if line.startswith(b";"):
                continue
            fields = line.decode("utf-8").strip().split()
            file_id = os.path.splitext(fields[0])[0]
            onset = float(fields[2])
            offset = float(fields[3])
            fid_to_score_regions[file_id].append((onset, offset))
    return UEM(fid_to_score_regions.items())



import numpy as np

from typing import List, Union, Optional

from dover_lap.libs.turn import Turn, merge_turns
from dover_lap.src.label_mapping import LabelMapping
from dover_lap.src.label_voting import LabelVoting


class DOVERLap:
    @classmethod
    def combine_turns_list(
        cls,
        turns_list: List[List[Turn]],
        file_id: str,
        label_mapping: Optional[str] = "greedy",
        second_maximal: Optional[bool] = False,
        voting_method: Optional[str] = "average",
        weight_type: Optional[str] = "rank",
        dover_weight: Optional[float] = 0.1,
        custom_weight: Optional[List[str]] = None,
        gaussian_filter_std: float = 0.01,
    ) -> List[List[Turn]]:

        # Label mapping stage
        mapped_turns_list, weights = LabelMapping.get_mapped_turns_list(
            turns_list,
            file_id,
            method=label_mapping,
            second_maximal=second_maximal,
        )

        # Compute weights based on rank
        if weight_type == "rank":
            ranks = cls.__get_ranks(weights)
            weights = cls.__compute_weights(ranks, dover_weight)
        elif weight_type == "custom":
            assert isinstance(custom_weight, list)
            weights = np.array([float(x) for x in custom_weight])
        elif weight_type == "norm":
            weights /= np.linalg.norm(weights, ord=1)  # use normalized weights

        # Label voting stage
        combined_turns_list = LabelVoting.get_combined_turns(
            mapped_turns_list,
            file_id,
            voting_method,
            weights,
            gaussian_filter_std=gaussian_filter_std,
        )
        # Merge consecutive turns with the same label
        combined_turns_list = merge_turns(combined_turns_list)
        return combined_turns_list

    def __get_ranks(weights: np.array) -> np.array:

        weights /= np.linalg.norm(weights, ord=1)
        temp = weights.argsort()
        ranks = np.empty_like(temp)
        ranks[temp] = np.arange(len(weights)) + 1
        return ranks

    def __compute_weights(ranks: np.array, weight: float) -> np.array:

        out_weights = 1 / np.power(ranks, weight)
        out_weights /= np.linalg.norm(out_weights, ord=1)
        return out_weights



"""Utility functions.

Taken from https://github.com/nryant/dscore
"""
import itertools
import sys
import click
import ast


def error(msg, file=sys.stderr):
    """Log error message ``msg`` to stderr."""
    msg = "ERROR: %s" % msg
    print(msg, file=file)


def info(msg, print_level=False, file=sys.stdout):
    """Log info message ``msg`` to stdout."""
    if print_level:
        msg = "INFO: %s" % msg
    print(msg, file=file)


def warn(msg, file=sys.stderr):
    """Log warning message ``msg`` to stderr."""
    msg = "WARNING: %s" % msg
    print(msg, file=file)


def xor(x, y):
    """Return truth value of ``x`` XOR ``y``."""
    return bool(x) != bool(y)


def format_float(x, n_digits=3):
    """Format floating point number for output as string.

    Parameters
    ----------
    x : float
        Number.

    n_digits : int, optional
        Number of decimal digits to round to.
        (Default: 3)

    Returns
    -------
    s : str
        Formatted string.
    """
    fmt_str = "%%.%df" % n_digits
    return fmt_str % round(x, n_digits)


def clip(x, lower, upper):
    """Clip ``x`` to [``lower``, ``upper``]."""
    return min(max(x, lower), upper)


def groupby(iterable, keyfunc):
    """Wrapper around ``itertools.groupby`` which sorts data first."""
    iterable = sorted(iterable, key=keyfunc)
    for key, group in itertools.groupby(iterable, keyfunc):
        yield key, group


# If an option is selected, other options become required
def command_required_option(require_name, require_map):
    class CommandOptionRequiredClass(click.Command):
        def invoke(self, ctx):
            require = ctx.params[require_name]
            if require not in require_map:
                raise click.ClickException(
                    "Unexpected value for --'{}': {}".format(require_name, require)
                )
            if (
                require_map[require] is not None
                and ctx.params[require_map[require].lower()] is None
            ):
                raise click.ClickException(
                    "With {}={} must specify option --{}".format(
                        require_name, require, require_map[require]
                    )
                )
            super(CommandOptionRequiredClass, self).invoke(ctx)

    return CommandOptionRequiredClass


# Class to accept list of arguments as Click option
class PythonLiteralOption(click.Option):
    def type_cast_value(self, ctx, value):
        try:
            return ast.literal_eval(value)
        except:
            return None



from typing import List, Dict, Optional

from dover_lap.libs.utils import groupby
from dover_lap.libs.turn import Turn

from dover_lap.src.mapping import HungarianMap, GreedyMap


class LabelMapping:
    @classmethod
    def get_mapped_turns_list(
        cls,
        turns_list: List[List[Turn]],
        file_id: str,
        method: Optional[str] = "greedy",
        second_maximal: Optional[bool] = False,
    ) -> List[List[Turn]]:
        """
        This function takes turns list from all RTTMs and applies an n-dimensional
        matching approximation algorithm to map all to a common label space.
        """

        if (len(turns_list) == 2) or (method == "hungarian"):
            # We replace the original turns list with one sorted by average DER
            hungarian_map = HungarianMap()
            label_mapping, weights = hungarian_map.compute_mapping(turns_list)
            turns_list = hungarian_map.sorted_turns_list

        elif method == "greedy":
            greedy_map = GreedyMap(second_maximal=second_maximal)
            label_mapping, weights = greedy_map.compute_mapping(turns_list)

        # Get mapped speaker labels using the mapping
        mapped_turns_list = []
        for i, turns in enumerate(turns_list):
            spk_groups = {
                key: list(group)
                for key, group in groupby(turns, lambda x: x.speaker_id)
            }
            mapped_turns = []
            for spk_id in spk_groups.keys():
                new_spk_id = label_mapping[(i, spk_id)]
                for turn in spk_groups[spk_id]:
                    mapped_turns.append(
                        Turn(
                            turn.onset,
                            turn.offset,
                            speaker_id=new_spk_id,
                            file_id=file_id,
                        )
                    )
            mapped_turns_list.append(mapped_turns)

        return mapped_turns_list, weights



import copy
import numpy as np

from collections import defaultdict
from typing import List, Optional, Tuple

from dover_lap.libs.turn import Token, Turn
from dover_lap.src.voting import WeightedAverageVoting


class LabelVoting:
    EPS = 1e-3  # to avoid float equality check errors

    @classmethod
    def get_combined_turns(
        cls,
        turns_list: List[List[Turn]],
        file_id: str,
        voting_method: Optional[str] = "average",
        weights: Optional[np.array] = None,
        gaussian_filter_std: float = 0.01,
    ) -> List[Turn]:
        """
        Implements combination using the DOVER-Lap label voting method.
        :param turns_list, list of mapped speaker turns (from each hypothesis)
        :param file_id, name of the file (recording/session)
        :param voting_method, which method to use for combining labels. Options:
            - `average`: use weighted average of labels
            - `hmm`: use a HMM-based voting method
        :param weights, hypothesis weights to use for rank weighting
        :param gaussian_filter_std, standard deviation of Gaussian filter to apply
        :return: combined turns
        """
        regions, start_end = cls.__get_regions(turns_list, weights)

        if voting_method == "average":
            voter = WeightedAverageVoting(gaussian_filter_std=gaussian_filter_std)
        else:
            raise ValueError("Unknown voting method: {}".format(voting_method))

        combined_turns_list = voter.get_combined_turns(regions, start_end, file_id)

        return combined_turns_list

    @classmethod
    def __get_regions(
        cls, turns_list: List[List[Turn]], weights: Optional[np.array] = None
    ) -> List[Tuple[float, float, List[Tuple[int, float]]]]:
        """
        Returns homogeneous time regions of the input.
        """
        # Map speaker ids to consecutive integers (0, 1, 2, ...)
        spk_index = {}
        for turns in turns_list:
            for turn in turns:
                if turn.speaker_id not in spk_index:
                    spk_index[turn.speaker_id] = len(spk_index)

        # Add weights to turns, and update speaker id. New speaker id is a tuple of (hyp_index, spk_index)
        if weights is None:
            weights = np.array([1] * len(turns_list))
        weighted_turns_list = []

        for i, (turns, weight) in enumerate(zip(turns_list, weights)):
            weighted_turns = [
                Turn(
                    turn.onset,
                    offset=turn.offset,
                    speaker_id=(i, spk_index[turn.speaker_id]),
                    file_id=turn.file_id,
                    weight=weight,
                )
                for turn in turns
            ]
            weighted_turns_list.append(weighted_turns)
        all_turns = [turn for turns in weighted_turns_list for turn in turns]

        tokens = []
        for turn in all_turns:
            # Name is 'START' (not 'BEG') so that 'END' tokens come first for same timestamp
            tokens.append(
                Token(
                    type="START",
                    timestamp=turn.onset,
                    hyp_index=turn.speaker_id[0],
                    speaker=turn.speaker_id[1],
                    weight=turn.weight,
                )
            )
            tokens.append(
                Token(
                    type="END",
                    timestamp=turn.offset,
                    hyp_index=turn.speaker_id[0],
                    speaker=turn.speaker_id[1],
                    weight=turn.weight,
                )
            )

        sorted_tokens = sorted(tokens, key=lambda x: (x.timestamp, x.type))

        regions_list = []
        region_start = sorted_tokens[0].timestamp
        # We also maintain a running dictionary of speakers and their weights contributed
        # by each hypothesis. These weights are stored as a list indexed by hyp_index.
        running_speakers_dict = defaultdict(
            lambda: [0.0 for _ in range(len(turns_list))]
        )
        running_speakers_dict[sorted_tokens[0].speaker][
            sorted_tokens[0].hyp_index
        ] += sorted_tokens[0].weight

        for token in sorted_tokens[1:]:
            if token.timestamp - region_start > cls.EPS:
                running_speakers = []
                for k, v in running_speakers_dict.items():
                    if sum(v) > cls.EPS:
                        running_speakers.append((k, v))
                if len(running_speakers) > 0:
                    regions_list.append(
                        (region_start, token.timestamp, copy.deepcopy(running_speakers))
                    )
            # Update the weights list for the current speaker
            weights = running_speakers_dict[token.speaker]
            if token.type == "START":
                weights[token.hyp_index] += token.weight
            else:
                weights[token.hyp_index] -= token.weight
            running_speakers_dict[token.speaker] = weights

            region_start = token.timestamp

        # Build regions matrix and start_end matrix. Regions matrix is of shape
        # (num_regions, num_speakers, num_hypotheses). start_end matrix is of shape
        # (num_regions, 2) and contains the start and end times of each region.
        regions = np.zeros(
            (len(regions_list), len(spk_index), len(turns_list)), dtype=np.float32
        )
        start_end = np.zeros((len(regions_list), 2), dtype=np.float32)
        for i, region in enumerate(regions_list):
            start_end[i, 0] = region[0]
            start_end[i, 1] = region[1]
            for spk, weights in region[2]:
                regions[i, spk] = weights

        return regions, start_end



from typing import List, Dict, Tuple

from dover_lap.libs.utils import groupby
from dover_lap.libs.turn import Turn


def compute_spk_overlap(ref_spk_turns: List[Turn], sys_spk_turns: List[Turn]) -> float:
    """
    Computes 'relative' overlap, i.e. Intersection Over Union
    """
    tokens = []
    all_turns = ref_spk_turns + sys_spk_turns
    for turn in all_turns:
        tokens.append(("BEG", turn.onset))
        tokens.append(("END", turn.offset))
    spk_count = 0
    ovl_duration = 0
    total_duration = 0
    for token in sorted(tokens, key=lambda x: x[1]):
        if token[0] == "BEG":
            spk_count += 1
            if spk_count == 2:
                ovl_begin = token[1]
            if spk_count == 1:
                speech_begin = token[1]
        else:
            spk_count -= 1
            if spk_count == 1:
                ovl_duration += token[1] - ovl_begin
            if spk_count == 0:
                total_duration += token[1] - speech_begin
    return ovl_duration / total_duration


def get_speaker_keys(turns_list: List[List[Turn]]) -> Dict[Tuple[int, int], str]:
    """
    Returns a dictionary which maps a file id (relative) and speaker id (relative)
    to absolute speaker id.
    """
    speakers_dict = {}
    for i, turns in enumerate(turns_list):
        turn_groups = {
            key: list(group) for key, group in groupby(turns, lambda x: x.speaker_id)
        }
        for j, key in enumerate(sorted(turn_groups.keys())):
            speakers_dict[(i, j)] = key
    return speakers_dict


import numpy as np
import sys

from typing import List, Dict, Tuple, Optional

from dover_lap.libs.utils import groupby, info
from dover_lap.libs.turn import Turn
from .map_utils import *


class GreedyMap:
    def __init__(self, second_maximal: Optional[bool] = False) -> None:
        self.second_maximal = second_maximal

    def compute_mapping(
        self,
        turns_list: List[List[Turn]],
    ) -> Tuple[Dict[Tuple[int, str], int], np.ndarray]:
        """
        Use the DOVER-Lap greedy label mapping algorithm. Returns a map from
        old labels to new, and the weights for the hypotheses.
        """
        self.turns_list = turns_list
        N = len(self.turns_list)
        cost_tensor, pairwise_costs = self.compute_cost_tensor(turns_list)

        # The weight of each hypothesis is computed by computing its total
        # overlap with all other hypotheses
        weights = np.array([0] * N, dtype=float)
        for i in range(N):
            cur_pairwise_costs = [
                np.squeeze(x) for x in pairwise_costs.values() if x.shape[i] != 1
            ]
            weights[i] = -1 * sum([np.sum(x) for x in cur_pairwise_costs])

        label_mapping = self._apply_maximal_matching(
            cost_tensor,
            get_speaker_keys(turns_list),
        )
        return (label_mapping, weights)

    def compute_cost_tensor(self, turns_list: List[List[Turn]]) -> np.ndarray:

        N = len(turns_list)
        k = int((N * (N - 1) / 2))
        pairwise_costs = {}

        has_single_speaker = False

        for i, ref_turns in enumerate(turns_list):
            for j, sys_turns in enumerate(turns_list):
                if j <= i:
                    continue
                cost = []
                ref_groups = {
                    key: list(group)
                    for key, group in groupby(ref_turns, lambda x: x.speaker_id)
                }
                sys_groups = {
                    key: list(group)
                    for key, group in groupby(sys_turns, lambda x: x.speaker_id)
                }

                if len(ref_groups.keys()) == 1 or len(sys_groups.keys()) == 1:
                    has_single_speaker = True

                for ref_spk_id in sorted(ref_groups.keys()):
                    cur_row = []
                    ref_spk_turns = ref_groups[ref_spk_id]
                    for sys_spk_id in sorted(sys_groups.keys()):
                        sys_spk_turns = sys_groups[sys_spk_id]
                        total_overlap = compute_spk_overlap(
                            ref_spk_turns, sys_spk_turns
                        )
                        cur_row.append(-1 * total_overlap)
                    cost.append(cur_row)

                new_axis = list(range(N))
                new_axis.remove(i)
                new_axis.remove(j)

                # The expand_dims is for easy broadcasting
                pairwise_costs[(i, j)] = np.expand_dims(
                    np.array(cost), axis=tuple(k for k in new_axis)
                )

        if has_single_speaker:
            # iterate and add since numpy cannot broadcast with 2 dummy dimensions
            vals = list(pairwise_costs.values())
            cost_tensor = vals[0]
            for val in vals[1:]:
                cost_tensor = np.add(cost_tensor, val)
        else:
            # otherwise use broadcasting
            cost_tensor = np.sum(list(pairwise_costs.values()))
        return cost_tensor, pairwise_costs

    def _apply_maximal_matching(
        self,
        cost_tensor: np.ndarray,
        speakers_dict: Dict[Tuple[int, int], str],
    ) -> List[List[Turn]]:

        # Sort the cost tensor
        sorted_idx = np.transpose(
            np.unravel_index(np.argsort(cost_tensor, axis=None), cost_tensor.shape)
        )

        # Get the maximal matching using an approximation algorithm
        M = []
        remaining_idx = {
            i: list(range(cost_tensor.shape[i])) for i in range(len(cost_tensor.shape))
        }

        iter = 1
        while len(remaining_idx.keys()) != 0:
            info(
                "Iteration {}: {} labels left to be mapped".format(
                    iter,
                    sum([len(v) for v in remaining_idx.values()]),
                ),
                file=sys.stderr,
            )
            sorted_idx_filtered = self._filter_sorted_index_list(
                sorted_idx, remaining_idx
            )

            # find initial maximal matching
            M_cur = []
            for idx in sorted_idx_filtered:
                if not self._contradicts(M_cur, idx):
                    M_cur.append(idx)

            if self.second_maximal:
                # find second maximal matching
                change = True
                while change:
                    change = False
                    for idx in list(M_cur):
                        M_cur.remove(idx)
                        M_r = self._find_remaining_maximal_matching(
                            M_cur, sorted_idx_filtered
                        )
                        if len(M_r) > 1:
                            M_cur = M_cur + M_r
                            change = True
                        else:
                            M_cur.append(idx)

            for idx in M_cur:
                for i, j in enumerate(idx):
                    if i in remaining_idx and j in remaining_idx[i]:
                        remaining_idx[i].remove(j)

            for i in list(remaining_idx.keys()):
                if len(remaining_idx[i]) == 0:
                    del remaining_idx[i]

            M += M_cur
            iter += 1

        label_mapping = {}
        for k, idx_tuple in enumerate(M):
            # For each speaker j in hypothesis i, assign new label k
            for i, j in enumerate(idx_tuple):
                old_spk_id = speakers_dict[(i, j)]
                if (i, old_spk_id) not in label_mapping:
                    label_mapping[(i, old_spk_id)] = k

        return label_mapping

    def _find_remaining_maximal_matching(
        self, M: List[Dict[int, int]], idx_list: List[Tuple[int, int]]
    ) -> List[Tuple[int, int]]:
        """
        Given a list of index tuples and a matching M, find a maximal
        matching on the "remaining" list, i.e., using only those index
        tuples which are not present in the original matching.
        """
        S_r = []
        for idx in list(idx_list):
            if not self._contradicts(M, idx):
                S_r.append(idx)

        M_r = []
        for idx in S_r:
            if not self._contradicts(M_r, idx):
                M_r.append(idx)

        return M_r

    def _filter_sorted_index_list(
        self, sorted_idx: List[np.ndarray], remaining_idx: List[Tuple[int, int]]
    ) -> List[np.ndarray]:
        """
        Filter the sorted list of index tuples to only retain tuples for which
        at least one element is in the remaining list.
        """
        sorted_idx_filtered = []
        for idx_tuple in sorted_idx:
            remaining = False
            for i, j in enumerate(idx_tuple):
                if i in remaining_idx and j in remaining_idx[i]:
                    remaining = True
                    break
            if remaining:
                sorted_idx_filtered.append(idx_tuple)
        return sorted_idx_filtered

    def _contradicts(self, M: List[Dict[int, int]], idx_tuple: List[int]) -> bool:
        """
        Check if an index tuple contradicts a matching, i.e. return True if
        any index in the tuple is already present in the matching.
        """
        for i, index in enumerate(idx_tuple):
            existing_idx = [idx[i] for idx in M]
            if index in existing_idx:
                return True
        return False



from typing import Optional, List

import numpy as np
from scipy.ndimage import gaussian_filter1d
from scipy.stats import rankdata

from dover_lap.libs.turn import Turn


class WeightedAverageVoting:
    def __init__(self, gaussian_filter_std: float = 0.01) -> None:
        self.gaussian_filter_std = gaussian_filter_std

    def get_combined_turns(
        self, regions: np.ndarray, start_end: np.ndarray, file_id: str
    ) -> List[Turn]:
        """
        Implements combination using the DOVER-Lap weighted average voting method.

        :param regions, matrix of shape (num_regions, num_speakers, num_hypothesis).
            The value in cell (t, k, n) represents the weight speaker `k` in region `t`
            contributed by hypothesis `n`.
        :param start_end, list of start and end times for each region
        """
        assert (
            regions.shape[0] == start_end.shape[0]
        ), "Regions and start_end must have the same number of rows"

        # Sum the weights from all hypotheses
        regions = np.sum(regions, axis=2)

        # Apply Gaussian filter to the regions matrix along the T axis
        regions = gaussian_filter1d(
            regions, sigma=self.gaussian_filter_std, axis=0, mode="nearest"
        )

        # Get the number of speakers in each region
        num_spks = np.sum(regions, axis=1).round().astype(int)

        # Rank the weights in each region. We use the min method to break ties. This
        # means that if two speakers have the same weight, they will be assigned the
        # rank of the lower speaker. Note that we negate the regions matrix because
        # rankdata() sorts in ascending order.
        spk_ranks_matrix = rankdata(-1 * regions, axis=1, method="min")

        # Create turns list by combining the regions and start_end
        combined_turns_list = []
        for i in range(len(spk_ranks_matrix)):
            start_time, end_time = start_end[i]
            spk_ranks = spk_ranks_matrix[i]
            for j, spk_rank in enumerate(spk_ranks):
                if spk_rank <= num_spks[i]:
                    turn = Turn(
                        onset=start_time,
                        offset=end_time,
                        speaker_id=j,
                        file_id=file_id,
                    )
                    combined_turns_list.append(turn)

        return combined_turns_list



"""Functions for reading/writing RTTM files.

Taken from https://github.com/nryant/dscore
"""
from .turn import Turn
from .utils import format_float


def _parse_rttm_line(line):
    line = line.decode("utf-8").strip()
    fields = line.split()
    if len(fields) < 9:
        raise IOError('Number of fields < 9. LINE: "%s"' % line)
    file_id = fields[1]
    speaker_id = fields[7]

    # Check valid turn onset.
    try:
        onset = float(fields[3])
    except ValueError:
        raise IOError('Turn onset not FLOAT. LINE: "%s"' % line)
    if onset < 0:
        raise IOError('Turn onset < 0 seconds. LINE: "%s"' % line)

    # Check valid turn duration.
    try:
        dur = float(fields[4])
    except ValueError:
        raise IOError('Turn duration not FLOAT. LINE: "%s"' % line)
    if dur <= 0:
        raise IOError('Turn duration <= 0 seconds. LINE: "%s"' % line)

    return Turn(onset, dur=dur, speaker_id=speaker_id, file_id=file_id)


def load_rttm(rttmf):
    """Load speaker turns from RTTM file.

    For a description of the RTTM format, consult Appendix A of the NIST RT-09
    evaluation plan.

    Parameters
    ----------
    rttmf : str
        Path to RTTM file.

    Returns
    -------
    turns : list of Turn
        Speaker turns.

    speaker_ids : set
        Speaker ids present in ``rttmf``.

    file_ids : set
        File ids present in ``rttmf``.

    References
    ----------
    NIST. (2009). The 2009 (RT-09) Rich Transcription Meeting Recognition
    Evaluation Plan. https://web.archive.org/web/20100606041157if_/http://www.itl.nist.gov/iad/mig/tests/rt/2009/docs/rt09-meeting-eval-plan-v2.pdf
    """
    with open(rttmf, "rb") as f:
        turns = []
        speaker_ids = set()
        file_ids = set()
        for line in f:
            if line.startswith(b"SPKR-INFO"):
                continue
            turn = _parse_rttm_line(line)
            turns.append(turn)
            speaker_ids.add(turn.speaker_id)
            file_ids.add(turn.file_id)
    return turns, speaker_ids, file_ids


def write_rttm(rttmf, turns, n_digits=3, channel=1):
    """Write speaker turns to RTTM file.

    For a description of the RTTM format, consult Appendix A of the NIST RT-09
    evaluation plan.

    Parameters
    ----------
    rttmf : str
        Path to output RTTM file.

    turns : list of Turn
        Speaker turns.

    n_digits : int, optional
        Number of decimal digits to round to.
        (Default: 3)

    channel : int, optional
        Output channel
        (Default: 1)

    References
    ----------
    NIST. (2009). The 2009 (RT-09) Rich Transcription Meeting Recognition
    Evaluation Plan. https://web.archive.org/web/20100606041157if_/http://www.itl.nist.gov/iad/mig/tests/rt/2009/docs/rt09-meeting-eval-plan-v2.pdf
    """
    with open(rttmf, "wb") as f:
        for turn in turns:
            fields = [
                "SPEAKER",
                turn.file_id,
                str(channel),
                format_float(turn.onset, n_digits),
                format_float(turn.dur, n_digits),
                "<NA>",
                "<NA>",
                str(turn.speaker_id),
                "<NA>",
                "<NA>",
            ]
            line = " ".join(fields)
            f.write(line.encode("utf-8"))
            f.write(b"\n")


def validate_rttm(rttmf):
    """Validate RTTM file.

    Parameters
    ----------
    rttmf : str
        Path to RTTM file.

    Returns
    -------
    file_ids : set of str
        File ids present in ``rttmf``.

    speaker_ids : set of str
        Speaker ids present in ``rttm``.

    error_messages : list of str
         Errors encountered in file.
    """
    with open(rttmf, "rb") as f:
        file_ids = set()
        speaker_ids = set()
        error_messages = []
        for line in f:
            if line.startswith(b"SPKR-INFO"):
                continue
            try:
                turn = _parse_rttm_line(line)
                file_ids.add(turn.file_id)
                speaker_ids.add(turn.speaker_id)
            except IOError as e:
                error_messages.append(e.args[0])
    return file_ids, speaker_ids, error_messages



#!/usr/bin/env python3
"""
 This is the official implementation for the DOVER-Lap algorithm. It combines
 overlap-aware diarization hypotheses to produce an output RTTM.

 Raj, D., García-Perera, L.P., Huang, Z., Watanabe, S., Povey, D., Stolcke, A., & Khudanpur, S. 
 DOVER-Lap: A Method for Combining Overlap-aware Diarization Outputs.
 IEEE Spoken Language Technology Workshop 2021.

 Copyright  2020  Desh Raj (Johns Hopkins University)
"""
import sys
import click
import random
import numpy as np

from typing import List

from dover_lap.libs.rttm import load_rttm, write_rttm
from dover_lap.libs.turn import merge_turns, trim_turns, Turn
from dover_lap.libs.uem import load_uem
from dover_lap.libs.utils import (
    error,
    info,
    warn,
    groupby,
    command_required_option,
    PythonLiteralOption,
)

from dover_lap.src.doverlap import DOVERLap


def load_rttms(rttm_list: List[str]) -> List[List[Turn]]:
    """Loads speaker turns from input RTTMs in a list of turns."""
    turns_list = []
    file_ids = []
    for rttm_fn in rttm_list:
        try:
            turns_, _, file_ids_ = load_rttm(rttm_fn)
            turns_list.append(turns_)
            file_ids += list(file_ids_)
        except IOError as e:
            error("Invalid RTTM file: %s. %s" % (rttm_fn, e))
            sys.exit(1)
    return turns_list


@click.argument("input_rttms", nargs=-1, type=click.Path(exists=True))
@click.argument("output_rttm", nargs=1, type=click.Path())
@click.option("-u", "--uem-file", type=click.Path(), help="UEM file path")
@click.option(
    "-c",
    "--channel",
    type=int,
    default=1,
    show_default=True,
    help="Use this value for output channel IDs",
)
@click.option("--random-seed", type=int, default=0)
@click.option(
    "--label-mapping",
    type=click.Choice(["hungarian", "greedy"]),
    default="greedy",
    show_default=True,
    help="Choose label mapping algorithm to use",
)
@click.option(
    "--second-maximal",
    is_flag=True,
    default=False,
    show_default=True,
    help="If this flag is set, run a second iteration of the maximal matching for"
    " greedy label mapping",
)
@click.option(
    "--voting-method",
    type=click.Choice(["average"]),
    default="average",
    show_default=True,
    help="Choose voting method to use:"
    " average: use weighted average to combine input RTTMs",
)
@click.option(
    "--weight-type",
    type=click.Choice(["rank", "custom", "norm"]),
    default="rank",
    help="Specify whether to use rank weighting or provide custom weights",
    show_default=True,
)
@click.option(
    "--dover-weight",
    type=float,
    default=0.1,
    help="DOVER weighting factor",
    show_default=True,
)
@click.option(
    "--custom-weight", cls=PythonLiteralOption, help="Weights for input RTTMs"
)
@click.option(
    "--gaussian-filter-std",
    type=float,
    default=0.5,
    help="Standard deviation for Gaussian filter applied before voting. This can help"
    " reduce the effect of outliers in the input RTTMs. For quick turn-taking, set"
    " this to a small value (e.g. 0.1). 0.5 is a good value for most cases. Set this"
    " to a very small value, e.g. 0.01, to remove filtering.",
    show_default=True,
)
@click.command(
    cls=command_required_option(
        "weight_type", {"custom": "custom_weight", "rank": "dover_weight", "norm": None}
    )
)
def main(
    input_rttms: List[click.Path],
    output_rttm: click.Path,
    uem_file: click.Path,
    channel: int,
    random_seed: int,
    **kwargs,  # these are passed directly to combine_turns_list() method
) -> None:
    """Apply the DOVER-Lap algorithm on the input RTTM files."""

    # Set random seeds globally
    random.seed(random_seed)
    np.random.seed(random_seed)

    # Load hypothesis speaker turns.
    info("Loading speaker turns from input RTTMs...", file=sys.stderr)
    turns_list = load_rttms(input_rttms)

    if uem_file is not None:
        info("Loading universal evaluation map...", file=sys.stderr)
        uem = load_uem(uem_file)

        # Trim turns to UEM scoring regions and merge any that overlap.
        info(
            "Trimming reference speaker turns to UEM scoring regions...",
            file=sys.stderr,
        )
        turns_list = [trim_turns(turns, uem) for turns in turns_list]

    info("Merging overlapping speaker turns...", file=sys.stderr)
    turns_list = [merge_turns(turns) for turns in turns_list]

    file_to_turns_list = dict()
    for turns in turns_list:
        for fid, g in groupby(turns, lambda x: x.file_id):
            if fid in file_to_turns_list:
                file_to_turns_list[fid].append(list(g))
            else:
                file_to_turns_list[fid] = [list(g)]

    # Run DOVER-Lap algorithm
    file_to_out_turns = dict()
    for file_id in file_to_turns_list:
        info("Processing file {}..".format(file_id), file=sys.stderr)
        turns_list = file_to_turns_list[file_id]
        random.shuffle(
            turns_list
        )  # We shuffle so that the hypothesis order is randomized
        file_to_out_turns[file_id] = DOVERLap.combine_turns_list(
            turns_list, file_id, **kwargs
        )

    # Write output RTTM file
    write_rttm(output_rttm, sum(list(file_to_out_turns.values()), []), channel=channel)



import os

from setuptools import find_packages, setup

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

long_description = open("README.md").read()

dev_requires = ["pre-commit", "black", "flake8"]

setup(
    name="dover-lap",
    version="1.3.0",
    author="Desh Raj",
    author_email="r.desh26@gmail.com",
    description="Combine overlap-aware diarization output RTTMs",
    keywords="diarization dover",
    url="https://github.com/desh2608/dover-lap",
    license="Apache-2.0 License",
    packages=find_packages(),
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
    entry_points={"console_scripts": ["dover-lap=dover_lap.dover_lap:main"]},
)
